#!/usr/bin/python3.6+
# -*- coding:utf-8 -*-
"""
@author: zhouwei
@date: 2023/06/15
@desc: minio对象存储封装
"""
import traceback
from typing import Union, IO, AnyStr
from os import PathLike

from yzcore.extensions.storage.base import StorageManagerBase, StorageRequestError, logger
from yzcore.extensions.storage.schemas import S3Config
from yzcore.extensions.storage.utils import AnyStr2BytesIO
from yzcore.extensions.storage.amazon.utils import wrap_request_return_bool, wrap_request_raise_404
from yzcore.utils import datetime2str

try:
    import boto3
    from boto3.session import Session
    from botocore.exceptions import ClientError
except ImportError:
    boto3 = None


class S3Manager(StorageManagerBase):

    def __init__(self, conf: S3Config):
        super(S3Manager, self).__init__(conf)
        self.endpoint_url = f'{self.scheme}://{self.endpoint}'

        self.__init()

    def __init(self, bucket_name=None):
        """初始化对象"""
        if boto3 is None:
            raise ImportError("'boto3' must be installed to use AmazonS3Manager")

        if bucket_name:
            self.bucket_name = bucket_name

        self.client = boto3.client(
            's3',
            aws_access_key_id=self.access_key_id,
            aws_secret_access_key=self.access_key_secret,
            endpoint_url=self.endpoint_url,
        )

    def create_bucket(self, bucket_name):
        pass

    def get_bucket_cors(self):
        pass

    def list_buckets(self):
        return self.client.list_buckets()['Buckets']

    @wrap_request_return_bool
    def is_exist_bucket(self, bucket_name=None):
        if bucket_name is None:
            bucket_name = self.bucket_name
        return self.client.head_bucket(Bucket=bucket_name)

    def delete_bucket(self, bucket_name=None):
        pass

    def get_sign_url(self, key, expire=0):
        url = self.client.generate_presigned_url(
            ClientMethod='get_object',
            Params={'Bucket': self.bucket_name, 'Key': key},
            ExpiresIn=expire or self.private_expire_time,
            HttpMethod='GET',
        )
        return '//' + url.split('//', 1)[-1]

    def post_sign_url(self, key):
        """"
        获取post上传文件时需要的参数
        :param key: 这里的key作为将来key的前缀
        :return: {'key': key, 'AWSAccessKeyId': '', 'policy': '', 'signature': ''}
        """
        return self.client.generate_presigned_post(
            Bucket=self.bucket_name,
            Key=key+'${filename}',
            ExpiresIn=self.policy_expire_time,
        )

    def put_sign_url(self, key):
        return self.client.generate_presigned_url(
            ClientMethod='put_object',
            Params={'Bucket': self.bucket_name, 'Key': key},
            ExpiresIn=self.policy_expire_time,
            HttpMethod='PUT',
        )

    def iter_objects(self, prefix='', delimiter='', max_keys=100, **kwargs):
        result = []
        for obj in self.client.list_objects_v2(Bucket=self.bucket_name, Prefix=prefix, Delimiter=delimiter, MaxKeys=max_keys).get('Contents', []):
            result.append({
                'key': obj.get('Key'),
                'url': self.get_file_url(obj.get('Key')),
                'size': obj.get('Size'),
            })
        return result

    @wrap_request_raise_404
    def get_object_meta(self, key: str):
        response = self.client.head_object(Bucket=self.bucket_name, Key=key)
        return {
            'etag': response['ETag'].strip('"').lower(),
            'size': response['ContentLength'],
            'last_modified': datetime2str(response['LastModified']),
            'content_type': response['ContentType'],
        }

    @wrap_request_raise_404
    def _set_object_headers(self, key: str, headers: dict):
        content_type = headers.pop('Content-Type')
        self.client.copy_object(
            Bucket=self.bucket_name,
            Key=key,
            CopySource={'Bucket': self.bucket_name, 'Key': key},
            ContentType=content_type,
            Metadata=headers,
            MetadataDirective='REPLACE',
        )
        return True

    @wrap_request_return_bool
    def file_exists(self, key):
        return self.client.head_object(Bucket=self.bucket_name, Key=key)

    @wrap_request_raise_404
    def download_stream(self, key, **kwargs):
        return self.client.get_object(Bucket=self.bucket_name, Key=key)['Body']

    @wrap_request_raise_404
    def download_file(self, key, local_name, **kwargs):
        self.client.download_file(Bucket=self.bucket_name, Key=key, Filename=local_name)

    def upload_file(self, filepath: Union[str, PathLike], key: str, **kwargs):
        """上传文件"""
        extra_args = {'ContentType': self.parse_content_type(key)}
        try:
            self.client.upload_file(Bucket=self.bucket_name, Key=key, Filename=filepath, ExtraArgs=extra_args)
            return self.get_file_url(key)
        except Exception:
            logger.error(f's3 upload error: {traceback.format_exc()}')
            raise StorageRequestError(f's3 upload error')

    def upload_obj(self, file_obj: Union[IO, AnyStr], key: str, **kwargs):
        """上传文件流"""
        extra_args = {'ContentType': self.parse_content_type(key)}
        try:
            if isinstance(file_obj, (str, bytes)):
                file_obj = AnyStr2BytesIO(file_obj)
            self.client.upload_fileobj(Bucket=self.bucket_name, Key=key, Fileobj=file_obj, ExtraArgs=extra_args)
            return self.get_file_url(key)
        except Exception:
            logger.error(f's3 upload error: {traceback.format_exc()}')
            raise StorageRequestError(f's3 upload error')

    def delete_object(self, key: str):
        """删除文件"""
        self.client.delete_object(Bucket=self.bucket_name, Key=key)
        return True

    def get_policy(
            self,
            filepath: str,
            callback_url: str,
            callback_data: dict,
            **kwargs
    ):
        form_data = self.post_sign_url(filepath)
        form_data['fields'].pop('key')
        data = {
            'mode': self.mode,
            'host': form_data['url'],
            'dir': filepath,
            'callback': {'url': callback_url, 'data': callback_data},
            **form_data['fields'],
        }
        return data

    @property
    def host(self):
        return self._host_minio

    def get_file_url(self, key, with_scheme=False):
        return self._get_file_url_minio(key, with_scheme)

    def get_key_from_url(self, url, urldecode=False):
        """从URL中获取对象存储key"""
        return self._get_key_from_url_minio(url, urldecode)



#!/usr/bin/python3.6+
# -*- coding:utf-8 -*-
"""
@author: zhouwei
@date: 2023/04/17
@desc: azure blob对象存储封装
"""
import traceback
from datetime import datetime, timedelta
from io import BytesIO
from typing import Union, IO, AnyStr
from os import PathLike

from yzcore.extensions.storage.base import StorageManagerBase, StorageRequestError, logger
from yzcore.extensions.storage.schemas import AzureConfig
from yzcore.extensions.storage.azure.utils import wrap_request_raise_404
from yzcore.utils.time_utils import datetime2str


try:
    from azure.storage.blob import BlobServiceClient, ContentSettings, ContainerClient, generate_blob_sas,\
        BlobSasPermissions
    from azure.core.exceptions import ResourceExistsError
except:
    BlobServiceClient = None


class AzureManager(StorageManagerBase):

    def __init__(self, conf: AzureConfig):
        super(AzureManager, self).__init__(conf)
        self.connection_string = conf.connection_string
        self.account_key = conf.account_key
        self.account_name = conf.account_name

        self.__init()

    def __init(self, bucket_name=None):
        """初始化对象"""

        if BlobServiceClient is None:
            raise ImportError("'azure-storage-blob' must be installed to use AzureManager")

        if bucket_name:
            self.bucket_name = bucket_name

        self.blob_service_client = BlobServiceClient.from_connection_string(self.connection_string)
        self.container_client = self.blob_service_client.get_container_client(self.bucket_name)

    def create_bucket(self, bucket_name):
        try:
            self.blob_service_client.create_container(bucket_name)
        except ResourceExistsError:
            pass
        self.bucket_name = bucket_name

    def get_bucket_cors(self):
        cors_dict = {
            'allowed_origins': [],
            'allowed_methods': [],
            'allowed_headers': [],
        }
        for cors_rule in self.blob_service_client.get_service_properties()['cors']:
            cors_dict['allowed_origins'] = cors_rule.allowed_origins
            cors_dict['allowed_methods'] = cors_rule.allowed_methods
            cors_dict['allowed_headers'] = cors_rule.allowed_headers
        return cors_dict

    def list_buckets(self):
        return self.blob_service_client.list_containers()

    def is_exist_bucket(self, bucket_name=None):
        if bucket_name is None:
            bucket_name = self.bucket_name
        container_client = self.blob_service_client.get_container_client(bucket_name)
        return container_client.exists()

    def delete_bucket(self, bucket_name=None):
        if bucket_name is None:
            bucket_name = self.bucket_name
        self.blob_service_client.delete_container(bucket_name)

    def get_sign_url(self, key, expire=0):
        expire_time = datetime.utcnow() + timedelta(seconds=expire or self.private_expire_time)
        blob_client = self.container_client.get_blob_client(blob=key)
        sas_sign = generate_blob_sas(
            account_name=self.account_name, container_name=self.bucket_name, blob_name=key, account_key=self.account_key,
            expiry=expire_time, permission=BlobSasPermissions(read=True)
        )
        url = f'{blob_client.url}?{sas_sign}'
        return '//' + url.split('//', 1)[-1]

    def post_sign_url(self, key):
        pass

    def put_sign_url(self, key):
        """
        获取put上传文件的链接
        请求必填的headers: x-ms-blob-type:BlockBlob
        文件直接放入 body：binary中
        """
        expire_time = datetime.utcnow() + timedelta(seconds=self.policy_expire_time)
        blob_client = self.container_client.get_blob_client(blob=key)
        sas_sign = generate_blob_sas(
            account_name=self.account_name, container_name=self.bucket_name, blob_name=key, account_key=self.account_key,
            expiry=expire_time, permission=BlobSasPermissions(write=True)
        )
        return f'{blob_client.url}?{sas_sign}'

    def get_file_url(self, key, with_scheme=False):
        return self._get_file_url_minio(key, with_scheme)

    @property
    def host(self):
        return self._host_minio

    def get_key_from_url(self, url, urldecode=False):
        """从URL中获取对象存储key"""
        return self._get_key_from_url_minio(url, urldecode)

    def iter_objects(self, prefix='', marker=None, delimiter=None, max_keys=100):
        objects = self.container_client.list_blobs(name_starts_with=prefix, results_per_page=max_keys)
        _result = []
        for obj in objects:
            _result.append({
                'key': obj.name,
                'url': self.get_file_url(obj.name),
                'size': obj.size,
            })
        return _result

    @wrap_request_raise_404
    def get_object_meta(self, key: str):
        """azure的etag不像 oss/obs/minio 是文件的md5，而content_md5需要在上传时指定"""
        blob_client = self.container_client.get_blob_client(blob=key)
        metadata = blob_client.get_blob_properties()
        content_md5 = metadata['content_settings']['content_md5'] or ''
        if isinstance(content_md5, bytearray):
            content_md5 = content_md5.hex()
        return {
            'etag': content_md5.lower(),  # metadata['etag'].strip('"').lower()
            'size': metadata['size'],
            'last_modified': datetime2str(metadata['last_modified']),
            'content_type': metadata['content_settings']['content_type']
        }

    @wrap_request_raise_404
    def _set_object_headers(self, key: str, headers: dict):
        blob_client = self.container_client.get_blob_client(blob=key)
        if not any([headers.get('content_md5'), headers.get('Content-MD5')]):
            metadata = blob_client.get_blob_properties()
            headers['content_md5'] = metadata['content_settings']['content_md5']
        blob_client.set_http_headers(ContentSettings(**headers))
        return True

    def file_exists(self, key):
        blob_client = self.container_client.get_blob_client(blob=key)
        return blob_client.exists()

    @wrap_request_raise_404
    def download_stream(self, key, **kwargs):
        blob_client = self.container_client.get_blob_client(blob=key)
        stream = BytesIO()
        blob_client.download_blob().readinto(stream)
        stream.seek(0)
        return stream

    @wrap_request_raise_404
    def download_file(self, key, local_name, **kwargs):
        blob_client = self.container_client.get_blob_client(blob=key)
        with open(local_name, 'wb') as f:
            f.write(blob_client.download_blob().readall())

    def upload_file(self, filepath: Union[str, PathLike], key: str, **kwargs):
        """上传文件流"""
        with open(filepath, 'rb') as f:
            return self.upload_obj(f, key)

    def upload_obj(self, file_obj: Union[IO, AnyStr], key: str, **kwargs):
        """上传文件流"""
        try:
            content_settings = ContentSettings(content_type=self.parse_content_type(key))
            blob_client = self.container_client.get_blob_client(blob=key)
            blob_client.upload_blob(file_obj, overwrite=True, content_settings=content_settings)
            return self.get_file_url(key)
        except Exception:
            logger.error(f'azure blob upload error: {traceback.format_exc()}')
            raise StorageRequestError(f'azure blob upload error')

    def delete_object(self, key: str):
        """删除文件"""
        blob_client = self.container_client.get_blob_client(blob=key)
        blob_client.delete_blob(delete_snapshots='include')
        return True

    def get_policy(
            self,
            filepath: str,
            callback_url: str,
            callback_data: dict,
            **kwargs
    ):
        """
        授权给第三方上传
        :param filepath: 需要在前一步拼接好完整的key
        :param callback_url: 对象存储的回调地址
        :param callback_data: 需要回传的参数
        :return:
        """
        return {
            'mode': self.mode,
            'host': self.put_sign_url(filepath),
            'headers': {'x-ms-blob-type': 'BlockBlob'},
            'callback': {'url': callback_url, 'data': callback_data},
        }



#!/usr/bin/python3.7+
# -*- coding:utf-8 -*-
"""
@auth: cml
@date: 2020-9-1
@desc: ...
"""
from yzcore.core.management.templates import TemplateCommand


class Command(TemplateCommand):
    help = (
        "Creates a Django app directory structure for the given app name in "
        "the current directory or optionally in the given directory."
    )
    missing_args_message = "You must provide an application name."

    def handle(self, **options):
        app_name = options.pop('name')
        target = options.pop('directory')
        super().handle('app', app_name, target, **options)




try:
    # >= python3.8
    from functools import cached_property
except ImportError:

    # copied from django.utils.functional.cached_property
    class cached_property:
        """
        Decorator that converts a method with a single self argument into a
        property cached on the instance.

        A cached property can be made out of an existing method:
        (e.g. ``url = cached_property(get_absolute_url)``).
        """

        name = None

        @staticmethod
        def func(instance):
            raise TypeError(
                "Cannot use cached_property instance without calling "
                "__set_name__() on it."
            )

        def __init__(self, func):
            self.real_func = func
            self.__doc__ = getattr(func, "__doc__")

        def __set_name__(self, owner, name):
            if self.name is None:
                self.name = name
                self.func = self.real_func
            elif name != self.name:
                raise TypeError(
                    "Cannot assign the same cached_property to two different names "
                    "(%r and %r)." % (self.name, name)
                )

        def __get__(self, instance, cls=None):
            """
            Call the function and put the return value in instance.__dict__ so that
            subsequent attribute access on the instance returns the cached value
            instead of calling cached_property.__get__().
            """
            if instance is None:
                return self
            res = instance.__dict__[self.name] = self.func(instance)
            return res


# copied from django.utils.functional.classproperty
class classproperty:
    """
    Decorator that converts a method with a single cls argument into a property
    that can be accessed directly from the class.
    """

    def __init__(self, method=None):
        self.fget = method

    def __get__(self, instance, cls=None):
        return self.fget(cls)

    def getter(self, method):
        self.fget = method
        return self



#!/usr/bin/python3.6+
# -*- coding:utf-8 -*-
"""
@auth: cml
@date: 2020-6-30
@desc: ...
"""
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, Union

try:
    from pydantic import BaseModel
    import sqlalchemy
    from sqlalchemy import text
    from sqlalchemy.orm import Session
    from sqlalchemy.ext.declarative import as_declarative, declared_attr
except ImportError:
    pass

from yzcore.core.encoders import jsonable_encoder


@as_declarative()
class Base:
    # Generate __tablename__ automatically
    @declared_attr
    def __tablename__(cls) -> str:
        return cls.__name__.lower()

    id: Any
    __name__: str


ModelType = TypeVar("ModelType", bound=Base)
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)


class OrmCRUDBase(Generic[ModelType, CreateSchemaType, UpdateSchemaType]):
    def __init__(self, model: Type[ModelType]):
        """
        CRUD object with default methods to Create, Read, Update, Delete (CRUD).
        **Parameters**
        * `model`: A SQLAlchemy model class
        * `schema`: A Pydantic model (schema) class
        """
        assert sqlalchemy is not None, "'sqlalchemy' must be installed to use OrmCRUDBase"
        self.model = model

    def count(self, db: Session, **kwargs):
        """
        根据条件获取总数量

        :param db:
        :param kwargs:
        :return:
        """
        if kwargs:
            return db.query(self.model).filter_by(**kwargs).count()
        return db.query(self.model).count()

    def get(self, db: Session, model_id: Any) -> Optional[ModelType]:
        """
        根据id获取数据

        :param db:
        :param model_id:
        :return:
        """
        return db.query(self.model).get(model_id)
    
    def get_one(self, db: Session, **kwargs):
        """
        根据查询条件获取一个数据
        
        :param db: 
        :param kwargs: 
        :return: 
        """
        return db.query(self.model).filter_by(**kwargs).one_or_none()
    
    def list(
            self, db: Session, *, sort: List[str] = None, offset: int = 0,
            limit: int = 100, **kwargs
    ) -> List[ModelType]:
        """
        根据查询条件获取数据列表

        :param db:
        :param sort: 需要排序的字段 ['-create_time', 'update_time'] (负号为降序)
        :param offset:
        :param limit:
        :param kwargs:
        :return:
        """
        if sort:
            sort = text(','.join(sort))
        if kwargs:
            return db.query(self.model).filter_by(
                **kwargs).order_by(sort).offset(offset).limit(limit).all()
        else:
            return db.query(self.model).order_by(sort).offset(
                offset).limit(limit).all()

    def create(
            self, db: Session, *,
            data: Union[Dict[str, Any], CreateSchemaType],
            is_transaction: bool = False
    ) -> ModelType:
        """
        插入操作，返回创建的模型

        :param db:
        :param data: 创建所需的数据
        :param is_transaction: 是否开启事务功能
        :return:
        """
        if isinstance(data, BaseModel):
            data = jsonable_encoder(data)
        db_obj = self.model(**data)  # type: ignore
        db.add(db_obj)
        if not is_transaction:
            db.commit()
            db.refresh(db_obj)
        return db_obj

    def update(
            self,
            db: Session, *,
            model_id: int = None,
            obj: ModelType = None,
            query: Dict[str, Any] = None,
            data: Union[UpdateSchemaType, Dict[str, Any]],
            is_return_obj: bool = False,
            is_transaction: bool = False

    ) -> ModelType:
        """
        单个对象更新

        更新有两种方式：
        方式一：
                传入需要更新的模型和更新的数据。
        方式二：
                传入id和更新的数据
        注意：
            如果传入模型来进行更新，则'is_return_obj=False'失效，返回更新后的模型

        :param db:
        :param model_id:    模型ID
        :param obj:         模型对象
        :param query:       模型查询参数
        :param data:        需要更新的数据
        :param is_return_obj:   是否需要返回模型数据，默认为False，只返回更新成功的行数
        :param is_transaction:  是否开启事务功能

        :return: update_count or obj or None
        """
        if not any((model_id, obj, query)):
            raise ValueError('At least one of [model_id、query、obj] exists')

        if isinstance(data, dict):
            update_data = data
        else:
            update_data = data.dict(exclude_unset=True)

        if not is_return_obj and not obj:
            if model_id:
                update_count = db.query(self.model).filter(
                    self.model.id == model_id).update(update_data)
            else:
                update_count = db.query(self.model).filter_by(
                    **query).update(update_data)

            if not is_transaction:
                db.commit()
            return update_count
        else:
            if not obj:
                if model_id:
                    obj = self.get(db, model_id)
                else:
                    obj = self.get_one(db, **query)
            if obj:
                obj_data = jsonable_encoder(obj)

                for field in obj_data:
                    if field in update_data:
                        setattr(obj, field, update_data[field])
                db.add(obj)
                if not is_transaction:
                    db.commit()
                    db.refresh(obj)
                return obj

    def delete(
            self, db: Session, *,
            model_id: int,
            is_return_obj: bool = False,
            is_transaction: bool = False
    ) -> ModelType:
        """

        :param db:
        :param model_id:      模型ID
        :param is_return_obj: 是否需要返回模型数据，默认为False，只返回删除成功的行数
        :param is_transaction:  是否开启事务功能
        :return:
        """
        if is_return_obj:
            obj = db.query(self.model).get(model_id)
            db.delete(obj)
            if not is_transaction:
                db.commit()
            return obj
        else:
            del_count = db.query(self.model).filter(
                self.model.id == model_id).delete(synchronize_session=False)
            if not is_transaction:
                db.commit()
            return del_count

    def bulk_delete(self, db: Session, ids: List[int] = None, **kwargs):
        """

        :param db:
        :param kwargs:
        :return:
        """
        if ids:
            del_count = db.query(self.model).filter(
                self.model.id.in_(ids)).delete(synchronize_session=False)
        else:
            del_count = db.query(self.model).filter_by(
                **kwargs).delete(synchronize_session=False)
        return del_count



#!/usr/bin/python3.7+
# -*- coding:utf-8 -*-
"""
@auth: cml
@date: 2021-1-18
@desc: 数据结构模块
"""
from typing import List, Tuple


class ValuesSortDict(dict):
    """
    多值排序字典
    每次value更新都会重新排序
    """
    def __init__(self, tuple_len_limit=4, tuple_default_value=('', 0, 0.0, 0.0)):
        """
        每个元组值的等长，默认为4
        :param tuple_len_limit:
        """
        self._len_limit = tuple_len_limit
        self._default_tuple = tuple_default_value
        super().__init__()

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, super().__repr__())

    def __getitem__(self, key):
        try:
            _values = super().__getitem__(key)
        except KeyError:
            _values = []
        return _values

    def __setitem__(self, key, value: List[Tuple]):
        if isinstance(value, list):
            super().__setitem__(key, value)
        elif isinstance(value, tuple) and len(value) == self._len_limit:
            super().__setitem__(key, [value])
        else:
            raise ValueError('This is not the value I want.')

    # def get_one(self, key, index=0, default=None):
    #     _values = self.__getitem__(key)
    #
    #     if val == []:
    #         return default
    #     return val

    def add(self, key, value: tuple):
        """

        :param key:
        :param value:
        :return:
        """
        if isinstance(value, str):
            v_list = [value]
            v_list.extend([v for i, v in enumerate(self._default_tuple) if i != 0])
            value = tuple(v_list)
        if isinstance(value, tuple):
            if len(value) != self._len_limit:
                raise ValueError(f'Error: len(tuple_value)!={self._len_limit}')
            _values = self.__getitem__(key)
            result = self._sort_values(value, _values)
            self.__setitem__(key, result)
        else:
            raise ValueError('This is not the value I want.')

    def _sort_values(self, value: Tuple, _values: List[Tuple]) -> List[Tuple]:
        """

        :param value:
        :param _values:
        :return:
        """
        # one
        # _values.append(value)
        # return sorted(_values, key=lambda x: x[1], reverse=False)

        # two 推荐
        for i, v in enumerate(_values):
            if value[1] < v[1]:
                _values.insert(i, value)
                return _values
        _values.append(value)
        return _values

    def increase(self, key, v, inc=1):
        """
        自增元组值的序位
        :param key: 键
        :param v: 元组值的第一位
        :param inc:
        :return:
        """

        _values = self.__getitem__(key)

        if isinstance(v, str):  # 推荐
            for i, _v in enumerate(_values):
                if _v[0] == v:
                    _v = _values.pop(i)
                    self.__setitem__(
                        key,
                        self._sort_values(
                            (
                                _v[0],
                                _v[1]+inc,
                                *[__v for i, __v in enumerate(_v) if i not in [0, 1]]
                            ),
                            _values
                        )
                    )
                    return
            # 如果遍历结束还未返回，报错处理
            raise ValueError(f'The value[0]:[{v}] is not exist')

        elif isinstance(v, tuple) and len(v) == self._len_limit:
            try:
                _values.remove(v)
            except ValueError:
                pass
            # v[1] += inc
            self.__setitem__(key, self._sort_values(
                (v[0], v[1]+inc,
                 *[_v for i, _v in enumerate(v) if i not in [0, 1]]),
                _values))
        else:
            raise ValueError(f'The value:[{v}] is error')


if __name__ == '__main__':
    import random
    # _values = [(f"ip{i}", random.randint(1, 10000)) for i in range(100000)]
    _values = [(f"ip{i}", i) for i in range(100000)]
    ipnum = random.randint(1, 100000)
    value = (f"ip{ipnum}", ipnum)
    print(value)
    import time

    d = ValuesSortDict()
    # d["model"] = _values
    d.add('model', value)
    start = time.time()
    d.increase("model", value)
    end1 = time.time()
    # print(d)
    d.increase("model", value[0])
    end2 = time.time()
    print("===>add:", end1-start)
    print("===>add0:", end2-end1)
    # print(d)

    # import pickle
    #
    # pickle.dump(q, file=open('history.pkl', 'wb'))  # 保存，注意使用二进制
    # q = pickle.load(file=open('history.pkl', 'rb'))  # 读取，注意使用二进制


from io import BytesIO
from typing import AnyStr
from urllib.parse import unquote, urlparse
from pathlib import Path
from yzcore.utils.crypto import get_random_string


def create_temp_file(text_length=16):
    """创建一个包含随机字符串的内存文件"""
    file = BytesIO(get_random_string(text_length).encode())
    return file


def get_url_path(url, urldecode=False):
    """提取URL中的path数据"""
    if not any([url.startswith('//'), url.startswith('http')]):
        url = '//' + url
    if urldecode:
        url = unquote(url)
    return urlparse(url).path


def get_filename(path):
    """从路径中提取文件名"""
    path = unquote(path)
    return Path(path).name


def AnyStr2BytesIO(t: AnyStr):
    """bytes, str to BytesIO"""
    if isinstance(t, bytes):
        obj = BytesIO(t)
    else:
        obj = BytesIO(t.encode())
    return obj



#!/usr/bin/python3.6+
# -*- coding:utf-8 -*-
"""
@author: zhouwei
@date: 2022/08/17
@desc: 华为云obs封装，依赖obs
"""
import base64
import json
from typing import Union, IO, AnyStr
from os import PathLike

from yzcore.extensions.storage.base import StorageManagerBase, StorageRequestError
from yzcore.extensions.storage.obs.utils import wrap_request_return_bool
from yzcore.extensions.storage.schemas import ObsConfig
from yzcore.exceptions import NotFoundObject

try:
    import obs
    from obs import SetObjectMetadataHeader
    from .obs_inherit import ObsClient
except:
    obs = None


class ObsManager(StorageManagerBase):

    def __init__(self, conf: ObsConfig):
        super(ObsManager, self).__init__(conf)
        self.callback_directly = conf.callback_directly

        self.__init()

    def __init(self, bucket_name=None):
        """初始化对象"""
        if obs is None:
            raise ImportError("'esdk-obs-python' must be installed to use ObsManager")

        if bucket_name:
            self.bucket_name = bucket_name

        # 创建ObsClient实例
        self.obsClient = ObsClient(
            access_key_id=self.access_key_id,
            secret_access_key=self.access_key_secret,
            server=self.endpoint,
        )

    def create_bucket(self, bucket_name=None, location='cn-south-1'):
        """创建bucket，并且作为当前操作bucket"""
        resp = self.obsClient.createBucket(bucket_name, location=location)
        if resp.status < 300:
            self.bucket_name = bucket_name
            return resp
        else:
            raise StorageRequestError(
                f"static_code: {resp.status}, errorCode: {resp.errorCode}. Message: {resp.errorMessage}.")

    def list_buckets(self):
        resp = self.obsClient.listBuckets(isQueryLocation=True)
        if resp.status < 300:
            return resp.body.buckets
        else:
            raise StorageRequestError(
                f"static_code: {resp.status}, errorCode: {resp.errorCode}. Message: {resp.errorMessage}.")

    @wrap_request_return_bool
    def is_exist_bucket(self, bucket_name=None):
        if bucket_name is None:
            bucket_name = self.bucket_name
        return self.obsClient.headBucket(bucket_name)

    def delete_bucket(self, bucket_name=None):
        if bucket_name is None:
            bucket_name = self.bucket_name
        return self.obsClient.deleteBucket(bucket_name)

    def get_sign_url(self, key, expire=0):
        res = self.obsClient.createSignedUrl(
            "GET", self.bucket_name, objectKey=key, expires=expire or self.private_expire_time)
        return '//' + res.signedUrl.split('//', 1)[-1]

    def post_sign_url(self, key, form_param=None):
        return self.obsClient.createPostSignature(
            self.bucket_name, objectKey=key, expires=self.policy_expire_time, formParams=form_param)

    def put_sign_url(self, key):
        res = self.obsClient.createSignedUrl(
            "PUT", self.bucket_name, objectKey=key, expires=self.policy_expire_time)
        return res.signedUrl

    def iter_objects(self, prefix='', marker=None, delimiter=None, max_keys=100):
        """
        遍历bucket下的文件
        :param prefix: key前缀
        :param marker:
        :param delimiter:
        :param max_keys:
        :return: dict
        """
        _result = []
        resp = self.obsClient.listObjects(self.bucket_name, prefix=prefix, marker=marker, delimiter=delimiter,
                                          max_keys=max_keys)
        if resp.status >= 300:
            raise StorageRequestError(
                f"static_code: {resp.status}, errorCode: {resp.errorCode}. Message: {resp.errorMessage}.")
        for obj in resp.body.contents:
            _result.append({
                'key': obj['key'],
                'url': self.get_file_url(key=obj['key']),
                'size': obj['size']
            })
        return _result

    def download_stream(self, key, **kwargs):
        resp = self.obsClient.getObject(self.bucket_name, key, loadStreamInMemory=False)
        if resp.status == 404:
            raise NotFoundObject()
        return resp.body.response

    def download_file(self, key, local_name, progress_callback=None):
        resp = self.obsClient.getObject(
            self.bucket_name, key,
            downloadPath=local_name,
            progressCallback=progress_callback
        )
        if resp.status == 404:
            raise NotFoundObject()

    def upload_file(self, filepath: Union[str, PathLike], key: str, **kwargs):
        """上传文件"""
        headers = obs.PutObjectHeader(contentType=self.parse_content_type(key))
        resp = self.obsClient.putFile(
            self.bucket_name, key, filepath, headers=headers)
        if resp.status >= 300:
            msg = resp.errorMessage
            raise StorageRequestError(f'obs upload error: {msg}')
        return self.get_file_url(key)

    def upload_obj(self, file_obj: Union[IO, AnyStr], key: str, **kwargs):
        """上传文件流"""
        headers = obs.PutObjectHeader(contentType=self.parse_content_type(key))
        resp = self.obsClient.putContent(
            self.bucket_name, key, content=file_obj, headers=headers)
        if resp.status >= 300:
            msg = resp.errorMessage
            raise StorageRequestError(f'obs upload error: {msg}')
        return self.get_file_url(key)

    def delete_object(self, key: str):
        """删除文件"""
        self.obsClient.deleteObject(self.bucket_name, key)
        return True

    def get_policy(
            self,
            filepath: str,
            callback_url: str,
            callback_data: dict,
            callback_content_type: str = "application/json",
    ):
        """
        授权给第三方上传
        由于华为云只有单az模式支持回调，多az不支持，可以根据不同情况选择obs回调或者前端发起回调
        callback_directly: True OBS直接发起回调 / False 由前端发起回调
        :param filepath:
        :param callback_url:
        :param callback_data: 需要回传的参数
        :param callback_content_type: 回调时的Content-Type
               "application/json"
               "application/x-www-form-urlencoded"
                华为云目前只能用application/json格式，用x-www-form-urlencoded时回调数据会在url中
        :return:
        """
        if self.callback_directly:
            callback_body = '{"filepath":"$(key)","etag":"$(etag)","size":$(fsize),"mime_type":"$(ext)",' \
                            '"data":' \
                            + json.dumps(callback_data) + '}'

            callback_body_plain = json.dumps(callback_body).strip().encode()
            base64_callback_body = base64.b64encode(callback_body_plain)

            form_param = {
                'body': base64_callback_body.decode(),
                'url': callback_url,
                'body-type': callback_content_type,
                # 'success_action_status': '200',
            }
        else:
            form_param = {}

        res = self.post_sign_url(key=filepath, form_param=form_param)

        data = dict(
            mode=self.mode,
            dir=filepath,
            host=f'{self.scheme}:{self.host}',
            AccessKeyId=self.access_key_id,
            policy=res.policy,
            signature=res.signature,
        )
        if not self.callback_directly:
            data['callback'] = {'url': callback_url, 'data': callback_data}
        return data

    def _set_object_headers(self, key: str, headers: dict):
        # 兼容 oss.update_file_headers
        obs_headers = SetObjectMetadataHeader()
        obs_headers.contentType = headers['Content-Type']  # oss 和 obs的参数名称不相同
        resp = self.obsClient.setObjectMetadata(self.bucket_name, key, headers=obs_headers)
        if resp.status == 404:
            raise NotFoundObject()
        return True

    @wrap_request_return_bool
    def file_exists(self, key):
        """检查文件是否存在"""
        return self.obsClient.headObject(self.bucket_name, key)

    def get_object_meta(self, key: str):
        """获取文件基本元信息，包括该Object的ETag、Size（文件大小）、LastModified，Content-Type，并不返回其内容"""
        resp = self.obsClient.getObjectMetadata(self.bucket_name, key)
        if resp.status == 404:
            raise NotFoundObject()
        return {
            'etag': resp.body.etag.strip('"').lower(),
            'size': resp.body.contentLength,
            'last_modified': resp.body.lastModified,
            'content_type': resp.body.contentType,
        }

    def get_bucket_cors(self):
        cors_dict = {
            'allowed_origins': [],
            'allowed_methods': [],
            'allowed_headers': [],
        }
        resp = self.obsClient.getBucketCors(self.bucket_name)
        if resp.status < 300:
            for rule in resp.body:
                cors_dict['allowed_origins'] = rule.allowedOrigin
                cors_dict['allowed_headers'] = rule.allowedHeader
                cors_dict['allowed_methods'] = rule.allowedMethod
            return cors_dict
        else:
            raise StorageRequestError(
                f"static_code: {resp.status}, errorCode: {resp.errorCode}. Message: {resp.errorMessage}.")



import functools
from yzcore.exceptions import NotFoundObject
try:
    from minio.error import S3Error
except ImportError:
    S3Error = None


def wrap_request_return_bool(func):
    """查询对象是否存在"""
    @functools.wraps(func)
    def wrap_func(*args, **kwargs):
        try:
            func(*args, **kwargs)
            return True
        except S3Error as exc:
            if exc.code != "NoSuchKey":
                raise
        return False
    return wrap_func


def wrap_request_raise_404(func):
    """对象不存在时抛出404"""
    @functools.wraps(func)
    def wrap_func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except S3Error as e:
            if e.code == 'NoSuchKey':
                raise NotFoundObject()
            raise
    return wrap_func



#!/usr/bin/python3.7+
# -*- coding:utf-8 -*-
"""
@auth: lxm
@date: 2020/12/14
@desc: ...
"""
import re

from _contextvars import Token
from starlette.requests import HTTPConnection
from starlette.responses import Response, PlainTextResponse
from starlette.types import ASGIApp, Receive, Scope, Send
from .fastapi_context import _request_scope_context_storage


class ContextMiddleware:
    def __init__(
            self,
            app: ASGIApp,
            default_error_response: Response = PlainTextResponse(content='缺少site-code请求头', status_code=412)
    ) -> None:
        self.app = app
        self.error_response = default_error_response

    async def __call__(
            self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        if scope["type"] not in ("http",):  # pragma: no cover
            await self.app(scope, receive, send)
            return
        request = HTTPConnection(scope)

        if scope['path'] == '/health':
            if re.match(r'^(127\.0\.0\.1)|(localhost)|(10\.\d{1,3}\.\d{1,3}\.\d{1,3})|(172\.((1[6-9])|(2\d)|(3[01]))\.\d{1,3}\.\d{1,3})|(192\.168\.\d{1,3}\.\d{1,3})$',request.client.host):
                await self.app(scope, receive, send)
                return

        site_code = request.headers.get('site-code', "")
        token: Token = _request_scope_context_storage.set({'site_code': site_code})

        if not site_code:
            message_head = {
                "type": "http.response.start",
                "status": self.error_response.status_code,
                "headers": self.error_response.raw_headers,
            }
            await send(message_head)

            message_body = {
                "type": "http.response.body",
                "body": self.error_response.body,
            }
            await send(message_body)
            return

        try:
            await self.app(scope, receive, send)
        finally:
            _request_scope_context_storage.reset(token)



import functools
from yzcore.exceptions import NotFoundObject
try:
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    ResourceNotFoundError = None


def wrap_request_raise_404(func):
    """对象不存在时抛出404"""
    @functools.wraps(func)
    def wrap_func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ResourceNotFoundError:
            raise NotFoundObject()
    return wrap_func



#!/usr/bin/python3.6+
# -*- coding:utf-8 -*-
"""
@auth: cml
@date: 2020-6-30
@desc: ...
"""
from urllib.parse import urlparse
from typing import Generator
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from yzcore.db.CustomQuery import Query
from yzcore.default_settings import default_setting as settings


def get_db_engine(uri=settings.DB_URI):
    if uri is None:
        raise EnvironmentError('需要配置"DB_URI"变量！')
    _typ = urlparse(uri).scheme
    if _typ.startswith('sqlite'):
        return create_engine(
            settings.DB_URI,
            connect_args={"check_same_thread": False}
            # 只有SQLite才需要，其他数据库不需要。SQLite 只允许一个线程与其通信
        )
    else:
        return create_engine(
            settings.DB_URI,
            connect_args={"connect_timeout": 5}
        )


engine = get_db_engine()
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine, query_cls=Query)

Base = declarative_base()


def get_db() -> Generator:
    try:
        db = SessionLocal()
        yield db
    finally:
        db.close()



#!/usr/bin/python3.7+
# -*- coding:utf-8 -*-
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="yz-core2",  # Replace with your own username
    version="1.0.61b1",
    author="zhouwe1",
    author_email="zhouwei@live.it",
    description="An ID generator for distributed microservices",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zhouwe1/yz-core.git",
    packages=setuptools.find_packages(),
    package_data={'yzcore': [
        'templates/project_template/*',
        'templates/project_template/docs/*',
        'templates/project_template/migrations/*',
        'templates/project_template/src/*',
        'templates/project_template/src/apps/*',
        'templates/project_template/src/conf/*',
        'templates/project_template/src/const/*',
        'templates/project_template/src/tests/*',
        'templates/project_template/src/utils/*',
        'templates/project_template/.gitignore',
    ]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6.8',
    install_requires=[
        "fastapi>=0.68.1",
        "uvicorn>0.13",
        "yzrpc>=0.1",
    ],
    entry_points={
        'console_scripts': [
            'yzcore=yzcore.__main__:main'
        ],
    },
    extras_require={
        'rpc': ['yzrpc>=0.1'],
        'storage-oss': ['oss2>=2.13.1, !=3.0.0'],
        'storage-obs': ['esdk-obs-python==3.22.2'],
        'storage-aws': ['boto3==1.26.157'],
        'storage-minio': ['minio==7.1.12'],
        'storage-azure': ['azure-storage-blob==12.16.0'],
    }
)
